/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef NUCLEICACIDUTILITIES_CPP
#define NUCLEICACIDUTILITIES_CPP

#include "nucleicAcidUtilities.h"

#include <stdexcept>
#include <string>
#include <limits.h>
#include <stdio.h>
#include <cmath>
#include <iostream>
#include "AssemblyException.h"
using namespace::std;


char* nucleicAcidUtilities::interpretString(char* rawSeq, long rawLen, set<char>* okToSkip){
  char* newSeqExtra = new char[rawLen];
  long n2 = 0;
  for (long n = 0; n < rawLen; ++n){
    n2++;
    switch(rawSeq[n]) {
    // allowed nucleotide characters
    case 'A': newSeqExtra[n2-1] = 'A'; break;
    case 'C': newSeqExtra[n2-1] = 'C'; break;
    case 'G': newSeqExtra[n2-1] = 'G'; break;
    case 'T': newSeqExtra[n2-1] = 'T'; break;
    case 'U': newSeqExtra[n2-1] = 'T'; break;
    case 'R': newSeqExtra[n2-1] = 'N'; break;
    case 'Y': newSeqExtra[n2-1] = 'N'; break;
    case 'S': newSeqExtra[n2-1] = 'N'; break;
    case 'W': newSeqExtra[n2-1] = 'N'; break;
    case 'K': newSeqExtra[n2-1] = 'N'; break;
    case 'M': newSeqExtra[n2-1] = 'N'; break;
    case 'B': newSeqExtra[n2-1] = 'N'; break;
    case 'D': newSeqExtra[n2-1] = 'N'; break;
    case 'H': newSeqExtra[n2-1] = 'N'; break;
    case 'V': newSeqExtra[n2-1] = 'N'; break;
    case 'N': newSeqExtra[n2-1] = 'N'; break;
    case 'a': newSeqExtra[n2-1] = 'A'; break;
    case 'c': newSeqExtra[n2-1] = 'C'; break;
    case 'g': newSeqExtra[n2-1] = 'G'; break;
    case 't': newSeqExtra[n2-1] = 'T'; break;
    case 'u': newSeqExtra[n2-1] = 'T'; break;
    case 'r': newSeqExtra[n2-1] = 'N'; break;
    case 'y': newSeqExtra[n2-1] = 'N'; break;
    case 's': newSeqExtra[n2-1] = 'N'; break;
    case 'w': newSeqExtra[n2-1] = 'N'; break;
    case 'k': newSeqExtra[n2-1] = 'N'; break;
    case 'm': newSeqExtra[n2-1] = 'N'; break;
    case 'b': newSeqExtra[n2-1] = 'N'; break;
    case 'd': newSeqExtra[n2-1] = 'N'; break;
    case 'h': newSeqExtra[n2-1] = 'N'; break;
    case 'v': newSeqExtra[n2-1] = 'N'; break;
    case 'n': newSeqExtra[n2-1] = 'N'; break;
    case '.': newSeqExtra[n2-1] = 'N'; break;
    default:
      if (okToSkip->count(rawSeq[n]) == 0){
	throw AssemblyException::ArgError("nucleotide seq contains invalid char");
      } else { n2--; }
    // no default; any other letter is skipped
    }
  }
  char* newSeqExtra2 = new char[n2+1];
  for (long n = 0; n < n2; ++n){ newSeqExtra2[n] = newSeqExtra[n]; }
  newSeqExtra2[n2] = '\0';
  delete [] newSeqExtra;
  return newSeqExtra2;
}



char nucleicAcidUtilities::reverseCompChar(char n){
  switch(n) {
  case 'A': return 'T';
  case 'C': return 'G';
  case 'G': return 'C';
  case 'T': return 'A';
  case 'N': return 'N';
  default:
    char errorMessage[100];
    char part1[] = "invalid nucleotide:";
    sprintf(errorMessage, "%s %c", part1, n);
    throw AssemblyException::ArgError(errorMessage);
  }
}


string nucleicAcidUtilities::reverseComplement(string seq){
  string rcSeq(seq.size(),' ');
  for (long n = 0; n < seq.size(); n++){
    rcSeq[seq.size() - n - 1] = reverseCompChar(seq[n]);
  }
  return rcSeq;
}
char* nucleicAcidUtilities::reverseComplement(char* seq, long length){
  char* rcSeq = new char[length+1];
  for (long n = 0; n < length; n++){
    rcSeq[length - n - 1] = reverseCompChar(seq[n]);
  }
  rcSeq[length] = 0; // the null ending that makes it a cstring
  return rcSeq;
}


/* KMER TOOLS
 * these are special tools/fields for dealing with kmers. note that,
 * for kmers, "N" will not be considered a valid nucleotide identity.
 */

// i use the natural log function to calculate the base-5 log of INT_MAX
long nucleicAcidUtilities::alphabetSize = 4; // <= 5 is the number of allowed letters (A,C,G,T)
int nucleicAcidUtilities::maxKmerSize = int(log( LONG_MAX ) / log( alphabetSize )); 
long nucleicAcidUtilities::maxKmerLong = 0;
void nucleicAcidUtilities::figureMaxKmerLong(){
  if ( maxKmerLong == 0 ){
    for ( int n = 0; n < maxKmerSize; n++ ){
      maxKmerLong += alphabetSize * long(pow(double(alphabetSize), n));
    }
  }
}


int nucleicAcidUtilities::getMaxKmerSize(){ return maxKmerSize; }
long nucleicAcidUtilities::getMaxKmerLong(){
  figureMaxKmerLong();
  return maxKmerLong;
}

void nucleicAcidUtilities::seqToSparseKmerLongs(set<long> * kmerSet, string * seq, int kmerSize){
  for (long n = 0; n < seq->size() - kmerSize + 1; n += kmerSize){
    try {
      kmerSet->insert( kmerToLong( seq->substr(n,kmerSize) ) );
    }
    catch (AssemblyException::KmerN_Error e){ }
    try {
      kmerSet->insert(kmerToLong( seq->substr( seq->size() - n - kmerSize, kmerSize ) ) );
    }
    catch (AssemblyException::KmerN_Error e){ }
  }
}


// an efficiency extension of the function below
// first nucs are at the low end; 0 -> empty string
// base-4: 1=A, 2=C, 3=G, 4=T (Ns are not valid kmer components)
void nucleicAcidUtilities::seqToKmerLongs(set<long> * kmerSet, string * seq, int kmerSize){
  if ( kmerSize > maxKmerSize ){
    throw AssemblyException::ArgError("nucAcidUtils:seqToCountedKmerLongs kmer size is greater than max kmer size");
  }

  long priorKmer = 0;
  for (long n = 0; n < seq->size() - kmerSize + 1; ++n){
    if ( priorKmer == 0 ){
      try { 
        priorKmer = kmerToLong( seq->substr(n,kmerSize) );
        kmerSet->insert(priorKmer);
      }
      catch (AssemblyException::KmerN_Error e){ }
    } else {
      try {
        long slidOver = (priorKmer - 1) / alphabetSize;
        int nucInt;
        char currentNuc = seq->at( n + kmerSize - 1 );

        switch(currentNuc){
        case 'A': nucInt = 1; break;
        case 'C': nucInt = 2; break;
        case 'G': nucInt = 3; break;
        case 'T': nucInt = 4; break;
        case 'N': throw AssemblyException::KmerN_Error("");
        default:
          char errorMessage[100];
          char part1[] = "invalid nucleotide:";
          sprintf(errorMessage, "%s %c", part1, currentNuc);
          throw AssemblyException::ArgError(errorMessage);
        }

        priorKmer = slidOver + nucInt * long(pow(double(alphabetSize), kmerSize - 1));
        kmerSet->insert(priorKmer);
      } catch (AssemblyException::KmerN_Error e){ priorKmer = 0; }
    }
  }
}



// a version of the function above that keeps track of how many times each kmer appeared
// in the sequence
void nucleicAcidUtilities::seqToCountedKmerLongs(map<long,long> * kmerSet, string * seq, int kmerSize, long edgeDistance){
  if ( kmerSize > maxKmerSize ){
    throw AssemblyException::ArgError("nucAcidUtils:seqToCountedKmerLongs kmer size is greater than max kmer size");
  }
  // these are both legal values
  long maxLowerWindow = edgeDistance - kmerSize;
  long minHigherWindow = seq->size() - edgeDistance;

  long priorKmer = 0;
  if (seq->size() >= kmerSize){
  for (long n = 0; n < seq->size() - kmerSize + 1; ++n){
    if (n <= maxLowerWindow or n >= minHigherWindow){ // imposes the edge constraints
      // note that this will nullify the the priorKmer trick at the point of transition
      // if and only if there is a jump in kmers (non-contiguous).  this is dealt with
      // below.

      if ( priorKmer == 0 ){
        try {
          priorKmer = kmerToLong( seq->substr(n,kmerSize) );
          map<long,long>::iterator kmerEntry = kmerSet->find( priorKmer );
          if ( kmerEntry == kmerSet->end() ){
            kmerSet->insert( pair<long,long> (priorKmer,1) );
          } else { kmerEntry->second++; }
        }
        catch (AssemblyException::KmerN_Error e){ priorKmer = 0; }
      } else {
        try {
          long slidOver = (priorKmer - 1) / alphabetSize;
          int nucInt;
          char currentNuc = seq->at( n + kmerSize - 1 );

          switch(currentNuc){
          case 'A': nucInt = 1; break;
          case 'C': nucInt = 2; break;
          case 'G': nucInt = 3; break;
          case 'T': nucInt = 4; break;
          case 'N': throw AssemblyException::KmerN_Error("");
          default:
            throw AssemblyException::ArgError("invalid nucleotide cannot be converted to long form");
          }

          priorKmer = slidOver + nucInt * long(pow(double(alphabetSize), kmerSize - 1));
          map<long,long>::iterator kmerEntry = kmerSet->find( priorKmer );
          if ( kmerEntry == kmerSet->end() ){
            kmerSet->insert( pair<long,long> (priorKmer,1) );
          } else { kmerEntry->second++; }
        } catch (AssemblyException::KmerN_Error e){ priorKmer = 0; }
      }
    } else { priorKmer = 0; } // else only executes if there was a skipped kmer
  }
  }
}


// first nucs are at the low end; 0 -> empty string
// base-4: 1=A, 2=C, 3=G, 4=T (Ns are not valid kmer components)
long nucleicAcidUtilities::kmerToLong(string kmer){
  if ( kmer.size() > maxKmerSize ){
    throw AssemblyException::ArgError("kmer is longer than max kmer size.");
  }
  // if the string is null, the intRep will remain zero
  long longRep = 0;
  int strRevPos = 0;
  string::iterator rit;
  for( rit = kmer.begin(); rit < kmer.end(); ++rit ){
    int nucInt;

    switch(*rit){
    case 'A': nucInt = 1; break;
    case 'C': nucInt = 2; break;
    case 'G': nucInt = 3; break;
    case 'T': nucInt = 4; break;
    case 'N': throw AssemblyException::KmerN_Error("");
    default:
      char errorMessage[100];
      char part1[] = "invalid nucleotide:";
      sprintf(errorMessage, "%s %c", part1, *rit);
      throw AssemblyException::ArgError(errorMessage);
    }

    longRep += nucInt * long(pow(double(alphabetSize), strRevPos));
    strRevPos++;
  }
  return longRep;
}

char nucleicAcidUtilities::numToChar[] = {'X','A','C','G','T'};

// must logically complement the kmerToInt function above
string nucleicAcidUtilities::longToKmer(long longRep){
  if ( longRep > getMaxKmerLong() ){
    throw AssemblyException::ArgError("this long is larger than the max kmer size.");
  }

  string strRep = "";
  long remainder;
  while ( longRep > 0 ){
    remainder = ((longRep - 1) % alphabetSize) + 1;
    longRep = (longRep - remainder) / alphabetSize;
    if (remainder > 4){ throw AssemblyException::LogicError("number is higher than any legit letter"); }
    else { strRep += numToChar[remainder]; }
  }
  return strRep;
}


  // These two are as derived by:
  // Jacquet & Szpankowski (1995) "Asymptotic behavior of the lempel-ziv parsing scheme and digital search trees"
  // Theoretical Computer Science 144.
/*
long nucleicAcidUtilities::predictedLzComplexity(long length){
  long double h = 1.3862943611198906; // this is the natural log of 4
  long double dubLen = long double(length);
  long double constant = h / log(dubLen);
  return long(dubLen * constant);
}
long nucleicAcidUtilities::predictedLzVariance(long length){
  // I calculated hsub2 to be 1.9218120556728056
  // since 
}
*/
long nucleicAcidUtilities::calculateLzComplexity(ScoredSeq* seq){
  // I will do this by constructing a tree, as is described in
  // Jacquet & Szpankowski (1995) "Asymptotic behavior of the lempel-ziv parsing scheme and digital search trees"
  // Theoretical Computer Science 144.
  LzTreeNode* fullTree = new LzTreeNode();
  long seqPos = 0;
  long wordCount = 0;
  long seqLen = seq->size();
  long wordStart = 0;
  bool wordIsOld = false;
  while (seqPos < seqLen){
    LzTreeNode* currentTree = fullTree;
    wordIsOld = true;
    while (wordIsOld and seqPos < seqLen){
      char nuc = seq->nucAtPosition(seqPos,'+');
      if (nuc != 'N'){
	LzTreeNode* branch = currentTree->getBranch(nuc);
	if (branch == 0){
	  wordIsOld = false;
	  currentTree->addBranch(nuc);
	  wordCount++;
	  wordStart = seqPos + 1;
	} else { currentTree = branch; }
      } else {
	if (seqPos==wordStart){ wordStart++; }
      }
      seqPos++;
    }
  }
  if (seqPos > wordStart){ wordCount++; }
  fullTree->deepDelete();
  return wordCount;
}


nucleicAcidUtilities::LzTreeNode::LzTreeNode(){
  for (int n = 0; n < 4; ++n){ _branches[n] = 0; }
}
nucleicAcidUtilities::LzTreeNode::~LzTreeNode(){}

void nucleicAcidUtilities::LzTreeNode::deepDelete(){
  for (int n = 0; n < 4; ++n){ if(_branches[n] != 0){ _branches[n]->deepDelete(); } }
  delete this;
}
nucleicAcidUtilities::LzTreeNode* nucleicAcidUtilities::LzTreeNode::getBranch(char nuc){
  int nucIndex;
  switch (nuc){
  case 'A': nucIndex = 0; break;
  case 'C': nucIndex = 1; break;
  case 'G': nucIndex = 2; break;
  case 'T': nucIndex = 3; break;
  default:
    throw AssemblyException::ArgError("invalid nuc input for LZW");
  }
  return _branches[nucIndex];
}
void nucleicAcidUtilities::LzTreeNode::addBranch(char nuc){
  int nucIndex;
  switch (nuc){
  case 'A': nucIndex = 0; break;
  case 'C': nucIndex = 1; break;
  case 'G': nucIndex = 2; break;
  case 'T': nucIndex = 3; break;
  default:
    throw AssemblyException::ArgError("invalid nuc input for LZW");
  }
  if (_branches[nucIndex] != 0){
    throw AssemblyException::CallingError("can't add a branch that already exists");
  } else {
    _branches[nucIndex] = new LzTreeNode();
  }
}






#endif
